;====================================================================
; NVIDIA NForce2 custom option ROM for ASUS A7N8X-E / 2.00 Deluxe
; vB.4 (bugfixing)
; author: digitalbath
; 10 dec 2024
; Credits to tictac, pinczakko, polygon, TzK and Infrared
; Version History
; v1.0: converted from ASUS A7N8X Option Rom, made by Tzk
; v2.0: added options from ASUS A7N8X Option Rom, made by Tzk
; v3.0: new build with code from Infrared
; vA.1: added alpha timings and romsips table
; vA.52: bugfixing, beta2
; vA.6: added alpha timings(M) and romsips table for Micron(O)
; vB.1: new Version
; vB.2A: added custom#6
; vB.4: new revision
;====================================================================
use16                                           ; 16bit mode
ROM_SIZE_IN_BLOCK = 12                          ; 1 means ROM size is 1 block (512 bytes)
ROM_SIZE_IN_BYTE = ROM_SIZE_IN_BLOCK * 512      ; number of 512 byte blocks

define PCI_ADDR_PORT    0xCF8
define PCI_DATA_PORT    0xCFC
define CMOS_ADDR_PORT   0x70
define CMOS_DATA_PORT   0x71

macro configurePort pci
{
    mov eax, pci                                ; move address for 32bit register offset
    mov dx, PCI_ADDR_PORT                       ; pci register address port
    out dx, eax                                 ; set port we want to read
}

macro readPciReg pci
{
    configurePort pci                           ; pci register address port
    mov dx, PCI_DATA_PORT                       ; pci register data port
    in eax, dx                                  ; read register values
}

macro writePciReg pci, data, mask
{
    readPciReg pci                              ; pci register data port
    and eax, mask                               ; mask for data to remove pre-set value
    or eax, data                                ; add custom value to old register data
    out dx, eax                                 ; write new custom register value
}

macro readCmosReg reg, mask
{
    mov al, reg                                 ; set CMOS index
    out CMOS_ADDR_PORT, al                      ; send register offset
    in al, CMOS_DATA_PORT                       ; read CMOS data
    and al, mask                                ; mask the data and zero unneeded bits
}

ROMStart:
    db 0x055, 0x0AA                             ; ROM Header 55,AA -> Bootable rom
    db (ROMEnd - ROMStart)/512                  ; ROM Size in 512 bytes
    jmp MAIN                                    ; jump to main routine
    db 0                                        ; checksum, to be filled in later
    times (256)-($-$$) db 0                     ; locate Main routine at 100h


MAIN:
    ; save all registers for later restore
    pushfd
    pusha
;
;====================================================================
;
; CMOS register and mask used in _EN_CODE.bin and system module:
;
; Cmos register 75h 0xF0h    Drive Strength  Bus: 0 Device: 0 Function: 4 Mem Controller
; Dimm 1 Drive Strength, Registers 7D, 81   8-bit
; Dimm 2 Drive Strength, Registers 67, 73   8-bit
; Dimm 3 Drive Strength, Registers 65, 71   8-bit
;
;
; Cmos register 75h 0x0Fh    Slew Rate Bus: 0 Device: 0 Function: 4 Mem Controller
; Dimm 1 Slew Rate, Register 7C 8-bit
; Dimm 2 Slew Rate, Register 66 8-bit
; Dimm 3 Slew Rate, Register 64 8-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 11h
;   2: 22h
;   3: 33h
;   4: 44h
;   5: 55h
;   6: 66h
;   7: 77h
;   8: 88h
;   9: 99h
;  10: AAh
;  11: BBh
;  12: CCh
;  13: DDh
;  14: EEh
;  15: FFh
;
;
; Cmos register 76h 0x1Fh  T(rc) Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 90h ?-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1T
;   ...
;   1F: 0001.1111b = 31T
;
;
; Cmos register 77h 0x1Fh T(rfc)  Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 91h ?-bit
;
;   0: Auto (keep value, skip custom code)
;   1: 0000.0001b = 1T
;   ...
;   1F: 0001.1111b = 31T
;
;
; Cmos register 74h 0x1Fh    alpha timings  Bus: 0 Device: 0 Function: 1 Mem Controller
;  Register 97h F0h 8-bit T(DOE)
;  Register 97h 0Fh 8-bit T(RRD)
;  Register 96h F0h 8-bit T(W2P)
;  Register 96h 0Fh 8-bit T(W2R)
;  Register 95h F0h 8-bit T(REXT)
;  Register 95h 0Fh 8-bit T(R2P)
;  Register 94h F0h 8-bit T(TR2W)
;
;   0: Auto (keep value, skip custom code)
;   1: A 0-0-3-3-1-0-3: 97h:00 96h:33 95h:10 94h:30
;   2: B 1-3-3-3-2-1-3: 97h:13 96h:33 95h:21 94h:30
;   3: C 1-3-3-3-3-3-4: 97h:13 96h:33 95h:33 94h:40
;   4: D 1-3-4-3-3-4-4: 97h:13 96h:43 95h:34 94h:40
;   5: E 2-3-3-3-3-3-4: 97h:23 96h:33 95h:33 94h:40
;   6: F 2-3-5-3-2-2-3: 97h:23 96h:53 95h:22 94h:30
;   7: G 3-2-5-3-2-2-3: 97h:32 96h:53 95h:22 94h:30
;   8: H 3-4-4-4-3-4-5: 97h:34 96h:44 95h:34 94h:50
;   9: I 3-4-5-4-3-4-5: 97h:34 96h:54 95h:34 94h:50
;   A: J 4-4-5-4-3-3-5: 97h:44 96h:54 95h:33 94h:50
;   B: K 4-5-5-5-3-5-6: 97h:45 96h:55 95h:35 94h:60
;   C: L 5-4-6-3-3-5-6: 97h:54 96h:63 95h:35 94h:60
;   D: M 5-5-6-5-3-5-6: 97h:55 96h:65 95h:35 94h:60
;   E: N 6-6-6-6-3-6-6: 97h:66 96h:66 95h:36 94h:60
;   F: O Winbond      : 97h:23 96h:53 95h:24 94h:50 /b0d0f1 A4h:02 B4:02 AC:02 F8 [2]=1 / bodof3= 48h:69 50h:B2 74h:62
; 
;
; Cmos register 78h 0x19h  T(ref)   Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 60h+61h 16-bit
;
;   00: Auto (keep value, skip custom code)
;   01: 0402 -> 0204h =  516 cycles
;   02: 8802 -> 0288h =  648 cycles
;   03: 0803 -> 0308h =  776 cycles
;   04: 0D03 -> 030Dh =  780 cycles
;   05: 8C03 -> 038Ch =  908 cycles
;   06: 0804 -> 0408h = 1032 cycles
;   07: 9004 -> 0490h = 1168 cycles
;   08: 1005 -> 0510h = 1296 cycles
;   09: 0006 -> 0600h = 1536 cycles
;   0A: 1006 -> 0610h = 1552 cycles
;   0B: 1806 -> 0618h = 1560 cycles
;   0C: 1807 -> 0718h = 1816 cycles
;   0D: 0008 -> 0800h = 2048 cycles
;   0E: 1008 -> 0810h = 2064 cycles
;   0F: 2009 -> 0920h = 2336 cycles
;   10: 000A -> 0A00h = 2560 cycles
;   11: 200A -> 0A20h = 2592 cycles
;   12: 000C -> 0C00h = 3072 cycles
;   13: 300C -> 0C30h = 3120 cycles
;   14: 300E -> 0E30h = 3632 cycles
;   15: 640E -> 0E64h = 3684 cycles
;   16: 2010 -> 1020h = 4128 cycles
;   17: 6410 -> 1064h = 4196 cycles
;   18: 4012 -> 1240h = 4672 cycles
;   19: 6412 -> 1264h = 4708 cycles
;
;
; Cmos register 73h 0x1Fh  romsips tables  Bus: 0 Device: 0 Function: 3 Mem Controller
;   8-bit Register 65h, 66h, 68h, 69h, 6Ah, 6Bh, 6Ch, 6Dh, 6Fh
;   0: Auto (keep value, skip custom code)
;   1: A DFI 12/10:  65h:06 66h:F7 68h:75 69h:00 6Ah:55 6Bh:00 6Ch:66 6Dh:55 6Fh:33
;   2: B DFI 12/18:  65h:06 66h:F7 68h:55 69h:00 6Ah:66 6Bh:00 6Ch:55 6Dh:55 6Fh:33
;   3: C DFI 12/31:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   4: D DFI 01/21:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:F0
;   5: E DFI 01/31:  65h:00 66h:F7 68h:33 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   6: F DFI 05/05:  65h:00 66h:F7 68h:55 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   7: G custom#1 :  65h:66 66h:F7 68h:22 69h:66 6Ah:55 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   8: H custom#2 :  65h:66 66h:77 68h:22 69h:07 6Ah:97 6Bh:00 6Ch:88 6Dh:88 6Fh:77
;   9: I custom#3 :  65h:44 66h:22 68h:22 69h:22 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:22
;   A: J custom#4 :  65h:44 66h:44 68h:44 69h:44 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:44
;   B: K custom#5 :  65h:66 66h:F7 68h:33 69h:66 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33   7Ch:44 48h:69 50h:B2
;   C: L custom#6 :  65h:66 66h:F7 68h:33 69h:66 6Ah:67 6Bh:00 6Ch:66 6Dh:66 6Fh:77   7Ch:44 48h:69 50h:B2
;   D:
;   E:
;   F:
;
;
; Cmos register 23h 0x1Fh  romsips tables  Bus: 0 Device: 0 Function: 3 Mem Controller
;   8-bit Register 68h
;
;   0: Auto (68h & 66h value will be changed by the FSB value, see below)
;   1: disabled (keep value, skip custom code)
;   2: 11h
;   3: 22h
;   4: 33h
;   5: 44h
;   6: 55h
;   7: 66h
;   8: 77h
;   9: 88h
;  10: 99h
;  11: AAh
;  12: BBh
;  13: CCh
;  14: DDh
;  15: EEh
;  16: FFh
; 
; auto FSB:  
;	FSB    |  66 + 68 Value  | CMOS 1B Value
; ===============================================
;  100-169MHz  |  F7 + FF	 | 0x - 3x
;  170-220MHz  |  F7 + DD        | 4x - 6x
;  221-238MHz  |  88 + 88        | 7x
;  239-254MHz  |  55 + 55        | 8x
;  255-269MHz  |  33 + 33        | 9x
;  270-300MHz  |  11 + 11        | Ax - Bx
;
;
;====================================================================

    readCmosReg 0x76, 0x1F                      ; read tRC from CMOS write to al

    cmp al, 0
    je trc_end                                  ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
    mov bl, al
    writePciReg 0x80000190, ebx, 0xFFFFFFE0     ;offset 90

    trc_end:					; jump target

;====================================================================

    readCmosReg 0x77, 0x1F                      ; read tRFC from CMOS

    cmp al, 0
    je trfc_end                                 ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
    mov bh, al
    writePciReg 0x80000190, ebx, 0xFFFFE0FF     ;offset 91

    trfc_end:                                   ; jump target

;====================================================================

    readCmosReg 0x75, 0x0F                      ; read slew rate from CMOS

    cmp al, 0
    je slewrate_end                             ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
    mov bl, al
    shl bl, 4
    or bl, al
    mov al, bl                                  ; save bl
    shl ebx, 16                                 ; shift bl to higher 16bit segment
    mov bl, al

    writePciReg 0x80000464, ebx, 0xFF00FF00     ;offset 64 and 66

    and ebx, 0xFF00FFFF                         ; discard the higher offset data

    writePciReg 0x8000047C, ebx, 0xFFFFFF00     ;offset 7C

    slewrate_end:                               ; jump target

;====================================================================
    
    readCmosReg 0x75, 0xF0                      ; read drive strength from CMOS

    cmp al, 0
    je drivestrength_end                        ; jump to next setting if auto selected (equal 0000.0000)

    xor ebx, ebx
    mov bl, al
    shr bl, 4
    or bl, al
    mov al, bl                                  ; save bl
    shl ebx, 24                                 ; shift bl to highest 8bit segment
    mov bh, al

    writePciReg 0x80000464, ebx, 0x00FF00FF     ; offset 65 and 67
    writePciReg 0x80000470, ebx, 0x00FF00FF     ; offset 71 and 73

    and ebx, 0x00FFFFFF

    writePciReg 0x8000047C, ebx, 0xFFFF00FF     ; offset 7D
    writePciReg 0x80000480, ebx, 0xFFFF00FF     ; offset 81

    drivestrength_end:                          ; jump target

;====================================================================

    readCmosReg 0x74, 0x1F                      ; read alpha from CMOS

    cmp al, 0
    je alpha_end                                ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .alpha1                                  ;jump if equal 0000.0001
    cmp al, 0x02
    je .alpha2                                  ;jump if equal 0000.0010
    cmp al, 0x03
    je .alpha3                                  ;jump if equal 0000.0011
    cmp al, 0x04
    je .alpha4                                  ;jump if equal 0000.0100
    cmp al, 0x05
    je .alpha5                                  ;jump if equal 0000.0101
    cmp al, 0x06
    je .alpha6                                  ;jump if equal 0000.0110
    cmp al, 0x07
    je .alpha7                                  ;jump if equal 0000.0111
    cmp al, 0x08
    je .alpha8                                  ;jump if equal 0000.1000
    cmp al, 0x09
    je .alpha9                                  ;jump if equal 0000.1001
    cmp al, 0x0A
    je .alpha10                                 ;jump if equal 0000.1010
    cmp al, 0x0B
    je .alpha11                                 ;jump if equal 0000.1011
    cmp al, 0x0C
    je .alpha12                                 ;jump if equal 0000.1100
    cmp al, 0x0D
    je .alpha13                                 ;jump if equal 0000.1101
    cmp al, 0x0E
    je .alpha14                                 ;jump if equal 0000.1110
    cmp al, 0x0F
    je .alpha15                                 ;jump if equal 0000.1111

  .alpha1:
    mov ebx, 0x00331030                         ;97h:00 96h:33 95h:10 94h:30
    je .end1
  .alpha2:
    mov ebx, 0x13332130                         ;97h:13 96h:33 95h:21 94h:30
    je .end1
  .alpha3:
    mov ebx, 0x13333340                         ;97h:13 96h:33 95h:33 94h:40
    je .end1
  .alpha4:
    mov ebx, 0x13433440                         ;97h:13 96h:43 95h:34 94h:40
    je .end1
  .alpha5:
    mov ebx, 0x23333340                         ;97h:23 96h:33 95h:33 94h:40
    je .end1
  .alpha6:
    mov ebx, 0x23532230                         ;97h:23 96h:53 95h:22 94h:30
    je .end1
  .alpha7:
    mov ebx, 0x32532230                         ;97h:32 96h:53 95h:22 94h:30
    je .end1
  .alpha8:
    mov ebx, 0x34443450                         ;97h:34 96h:44 95h:34 94h:50
    je .end1
  .alpha9:
    mov ebx, 0x34543450                         ;97h:34 96h:54 95h:34 94h:50
    je .end1
  .alpha10:
    mov ebx, 0x44543350                         ;97h:44 96h:54 95h:33 94h:50
    je .end1
  .alpha11:
    mov ebx, 0x45553560                         ;97h:45 96h:55 95h:35 94h:60
    je .end1
  .alpha12:
    mov ebx, 0x54633560                         ;97h:54 96h:63 95h:35 94h:60
    je .end1
  .alpha13:
    mov ebx, 0x55653560                         ;97h:55 96h:65 95h:35 94h:60
    je .end1
  .alpha14:
    mov ebx, 0x66663660                         ;97h:66 96h:66 95h:36 94h:60
    je .end1
  .alpha15:
    writePciReg 0x800001A4, 0x00000002, 0xFFFFFFF0     ;A4h:02
    writePciReg 0x800001B4, 0x00000002, 0xFFFFFFF0     ;B4h:02
    writePciReg 0x800001AC, 0x00000002, 0xFFFFFFF0     ;ACh:02
    writePciReg 0x800001F8, 0x00000004, 0xFFFFFFFB     ;F8h [2]=1
    writePciReg 0x80000348, 0x00000069, 0xFFFFFF00     ;48h:69
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ;50h:B2 
    writePciReg 0x80000374, 0x00000062, 0xFFFFFF00     ;74h:62
    mov ebx, 0x23532450                         ;97h:23 96h:53 95h:24 94h:50
    je .end1

  .end1:
    writePciReg 0x80000194, ebx, 0x0000000F
    je alpha_end

    alpha_end:                                  ; jump target


;====================================================================

    readCmosReg 0x78, 0x1F                      ; read tREF from CMOS

    cmp al, 0
    je tref_end                                ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .tref1                                   ;jump if equal 0000.0001
    cmp al, 0x02
    je .tref2                                   ;jump if equal 0000.0010
    cmp al, 0x03
    je .tref3                                   ;jump if equal 0000.0011
    cmp al, 0x04
    je .tref4                                   ;jump if equal 0000.0100
    cmp al, 0x05
    je .tref5                                   ;jump if equal 0000.0101
    cmp al, 0x06
    je .tref6                                   ;jump if equal 0000.0110
    cmp al, 0x07
    je .tref7                                   ;jump if equal 0000.0111
    cmp al, 0x08
    je .tref8                                   ;jump if equal 0000.1000
    cmp al, 0x09
    je .tref9                                   ;jump if equal 0000.1001
    cmp al, 0x0A
    je .tref10                                  ;jump if equal 0000.1010
    cmp al, 0x0B
    je .tref11                                  ;jump if equal 0000.1011
    cmp al, 0x0C
    je .tref12                                  ;jump if equal 0000.1100
    cmp al, 0x0D
    je .tref13                                  ;jump if equal 0000.1101
    cmp al, 0x0E
    je .tref14                                  ;jump if equal 0000.1110
    cmp al, 0x0F
    je .tref15                                  ;jump if equal 0000.1111
    cmp al, 0x10
    je .tref16                                  ;jump if equal 0001.0000
    cmp al, 0x11
    je .tref17                                  ;jump if equal 0001.0001
    cmp al, 0x12
    je .tref18                                  ;jump if equal 0001.0010
    cmp al, 0x13
    je .tref19                                  ;jump if equal 0001.0011
    cmp al, 0x14
    je .tref20                                  ;jump if equal 0001.0100
    cmp al, 0x15
    je .tref21                                  ;jump if equal 0001.0101
    cmp al, 0x16
    je .tref22                                  ;jump if equal 0001.0110
    cmp al, 0x17
    je .tref23                                  ;jump if equal 0001.0111
    cmp al, 0x18
    je .tref24                                  ;jump if equal 0001.1000
    cmp al, 0x19
    je .tref25                                  ;jump if equal 0001.1001
    

  .tref1:
    mov ebx, 0x0204                             ; 0516 Cycles
    je .end2
  .tref2:
    mov ebx, 0x0288                             ; 0648 Cycles
    je .end2
  .tref3:
    mov ebx, 0x0308                             ; 0776 Cycles
    je .end2
  .tref4:
    mov ebx, 0x030C                             ; 0780 Cycles
    je .end2
  .tref5:
    mov ebx, 0x038C                             ; 0908 Cycles
    je .end2
  .tref6:
    mov ebx, 0x0408                             ; 1032 Cycles
    je .end2
  .tref7:
    mov ebx, 0x0490                             ; 1168 Cycles
    je .end2
  .tref8:
    mov ebx, 0x0510                             ; 1296 Cycles
    je .end2
  .tref9:
    mov ebx, 0x0600                             ; 1536 Cycles
    je .end2
  .tref10:
    mov ebx, 0x0610                             ; 1552 Cycles
    je .end2
  .tref11:
    mov ebx, 0x0618                             ; 1560 Cycles
    je .end2
  .tref12:
    mov ebx, 0x0718                             ; 1816 Cycles
    je .end2
  .tref13:
    mov ebx, 0x0800                             ; 2048 Cycles
    je .end2
  .tref14:
    mov ebx, 0x0810                             ; 2064 Cycles
    je .end2
  .tref15:
    mov ebx, 0x0920                             ; 2336 Cycles
    je .end2
  .tref16:
    mov ebx, 0x0A00                             ; 2560 Cycles
    je .end2
  .tref17:
    mov ebx, 0x0A20                             ; 2592 Cycles
    je .end2
  .tref18:
    mov ebx, 0x0C00                             ; 3072 Cycles
    je .end2
  .tref19:
    mov ebx, 0x0C30                             ; 3120 Cycles
    je .end2
  .tref20:
    mov ebx, 0x0E30                             ; 3632 Cycles
    je .end2
  .tref21:
    mov ebx, 0x0E64                             ; 3684 Cycles
    je .end2
  .tref22:
    mov ebx, 0x1020                             ; 4128 Cycles
    je .end2
  .tref23:
    mov ebx, 0x1064                             ; 4196 Cycles
    je .end2
  .tref24:
    mov ebx, 0x1240                             ; 4672 Cycles
    je .end2
  .tref25:
    mov ebx, 0x1264                             ; 4708 Cycles
    je .end2

  .end2:
    writePciReg 0x80000160, ebx, 0xFFFF0000
    je tref_end

    tref_end:                                  ; jump target

;====================================================================

    readCmosReg 0x73, 0x1F                      ; read romsips from CMOS

    cmp al, 0
    je codeend                                  ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .sips1                                   ;jump if equal 0000.0001
    cmp al, 0x02
    je .sips2                                   ;jump if equal 0000.0010
    cmp al, 0x03
    je .sips3                                   ;jump if equal 0000.0011
    cmp al, 0x04
    je .sips4                                   ;jump if equal 0000.0100
    cmp al, 0x05
    je .sips5                                   ;jump if equal 0000.0101
    cmp al, 0x06
    je .sips6                                   ;jump if equal 0000.0110
    cmp al, 0x07
    je .sips7                                   ;jump if equal 0000.0111
    cmp al, 0x08
    je .sips8                                   ;jump if equal 0000.1000
    cmp al, 0x09
    je .sips9                                   ;jump if equal 0000.1001
    cmp al, 0x0A
    je .sips10                                  ;jump if equal 0000.1010
    cmp al, 0x0B
    je .sips11                                  ;jump if equal 0000.1011
    cmp al, 0x0C
    je .sips12                                  ;jump if equal 0000.1100
    cmp al, 0x0D
    je .sips13                                  ;jump if equal 0000.1101
    cmp al, 0x0E
    je .sips14                                  ;jump if equal 0000.1110
    cmp al, 0x0F
    je .sips15                                  ;jump if equal 0000.1111

  .sips1:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00550075, 0x00000000     ;6Bh:00 6Ah:55 69h:00 68h:75   
    writePciReg 0x8000036C, 0x33005566, 0x00FF0000     ;6Fh:33        6Dh:55 6Ch:66  
    jmp .sip_end
  .sips2:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660055, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:55   
    writePciReg 0x8000036C, 0x33005555, 0x00FF0000     ;6Fh:33        6Dh:55 6Ch:55  
    jmp .sip_end
  .sips3:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660066, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:66   
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66  
    jmp .sip_end
  .sips4:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660066, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:66   
    writePciReg 0x8000036C, 0xF0006666, 0x00FF0000     ;6Fh:F0        6Dh:66 6Ch:66  
    jmp .sip_end
  .sips5:
    writePciReg 0x80000364, 0x00F70000, 0xFF0000FF     ;       66h:F7 65h:00
    writePciReg 0x80000368, 0x00550633, 0x00000000     ;6Bh:00 6Ah:55 69h:06 68h:33   
    writePciReg 0x8000036C, 0x00000066, 0x00FF0000     ;6Fh:00        6Dh:00 6Ch:66  
    jmp .sip_end
  .sips6:
    writePciReg 0x80000364, 0x00F70000, 0xFF0000FF     ;       66h:F7 65h:00
    writePciReg 0x80000368, 0x00550655, 0x00000000     ;6Bh:00 6Ah:55 69h:06 68h:55   
    writePciReg 0x8000036C, 0x00000066, 0x00FF0000     ;6Fh:00        6Dh:00 6Ch:66  
    jmp .sip_end
  .sips7:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66
    writePciReg 0x80000368, 0x00556622, 0x00000000     ;6Bh:00 6Ah:55 69h:66 68h:22   
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66  
    jmp .sip_end
  .sips8:
    writePciReg 0x80000364, 0x00776600, 0xFF0000FF     ;       66h:77 65h:66
    writePciReg 0x80000368, 0x00970722, 0x00000000     ;6Bh:00 6Ah:97 69h:07 68h:22   
    writePciReg 0x8000036C, 0x77008888, 0x00FF0000     ;6Fh:77        6Dh:88 6Ch:88  
    jmp .sip_end
  .sips9:
    writePciReg 0x80000364, 0x00224400, 0xFF0000FF     ;       66h:22 65h:44
    writePciReg 0x80000368, 0x00442222, 0x00000000     ;6Bh:00 6Ah:44 69h:22 68h:22   
    writePciReg 0x8000036C, 0x22004444, 0x00FF0000     ;6Fh:22        6Dh:44 6Ch:44  
    jmp .sip_end
  .sips10:
    writePciReg 0x80000364, 0x00444400, 0xFF0000FF     ;       66h:44 65h:44
    writePciReg 0x80000368, 0x00444444, 0x00000000     ;6Bh:00 6Ah:44 69h:44 68h:44   
    writePciReg 0x8000036C, 0x44004444, 0x00FF0000     ;6Fh:44        6Dh:44 6Ch:44  
    jmp .sip_end
  .sips11:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66  
    writePciReg 0x80000368, 0x00666633, 0x00000000     ;6Bh:00 6Ah:66 69h:66 68h:33     
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66 
    writePciReg 0x8000037C, 0x00000044, 0xFFFFFF00     ;                     7Ch:44
    writePciReg 0x80000348, 0x00000069, 0xFFFFFF00     ;                     48h:69
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ;                     50h:B2
    jmp .sip_end
  .sips12:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66  
    writePciReg 0x80000368, 0x00676633, 0x00000000     ;6Bh:00 6Ah:67 69h:66 68h:33     
    writePciReg 0x8000036C, 0x77006666, 0x00FF0000     ;6Fh:77        6Dh:66 6Ch:66
    writePciReg 0x8000037C, 0x00000044, 0xFFFFFF00     ;                     7Ch:44
    writePciReg 0x80000348, 0x00000069, 0xFFFFFF00     ;                     48h:69
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ;                     50h:B2                                                 ;6Fh:          6Dh:   6Ch:    
    jmp .sip_end
  .sips13:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h:     
;                                                      ;6Fh:          6Dh:   6Ch:    
    jmp .sip_end
  .sips14:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h:     
;                                                      ;6Fh:          6Dh:   6Ch:    
    jmp .sip_end
  .sips15:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h: 
;                                                      ;6Fh:          6Dh:   6Ch:      
  .sip_end:
;
;  
;====================================================================
;
    readCmosReg 0x23, 0x1F                             ; read sip_68h from CMOS write to al
;
    cmp al, 0x01
    je codeend   
;
    xor ebx, ebx
;
    cmp al, 0
    je .sixtyeighth0                                   ;jump if equal 0000.0000
    cmp al, 0x02
    je .sixtyeighth2                                   ;jump if equal 0000.0010
    cmp al, 0x03
    je .sixtyeighth3                                   ;jump if equal 0000.0011
    cmp al, 0x04
    je .sixtyeighth4                                   ;jump if equal 0000.0100
    cmp al, 0x05
    je .sixtyeighth5                                   ;jump if equal 0000.0101
    cmp al, 0x06
    je .sixtyeighth6                                   ;jump if equal 0000.0110
    cmp al, 0x07
    je .sixtyeighth7                                   ;jump if equal 0000.0111
    cmp al, 0x08
    je .sixtyeighth8                                   ;jump if equal 0000.1000
    cmp al, 0x09
    je .sixtyeighth9                                   ;jump if equal 0000.1001
    cmp al, 0x0A
    je .sixtyeighth10                                  ;jump if equal 0000.1010
    cmp al, 0x0B
    je .sixtyeighth11                                  ;jump if equal 0000.1011
    cmp al, 0x0C
    je .sixtyeighth12                                  ;jump if equal 0000.1100
    cmp al, 0x0D
    je .sixtyeighth13                                  ;jump if equal 0000.1101
    cmp al, 0x0E 
    je .sixtyeighth14                                  ;jump if equal 0000.1110
    cmp al, 0x0F
    je .sixtyeighth15                                  ;jump if equal 0000.1111
    cmp al, 0x10
    je .sixtyeighth16                                  ;jump if equal 0001.0000
;
;

  .sixtyeighth2:
    writePciReg 0x80000368, 0x00000011, 0xFFFFFF00     ;68h:11
    jmp codeend 
  .sixtyeighth3:
    writePciReg 0x80000368, 0x00000022, 0xFFFFFF00     ;68h:22  
    jmp codeend 
  .sixtyeighth4:
    writePciReg 0x80000368, 0x00000033, 0xFFFFFF00     ;68h:33 
    jmp codeend 
  .sixtyeighth5:
    writePciReg 0x80000368, 0x00000044, 0xFFFFFF00     ;68h:44 
    jmp codeend 
  .sixtyeighth6:
    writePciReg 0x80000368, 0x00000055, 0xFFFFFF00     ;68h:55 
    jmp codeend 
  .sixtyeighth7:
    writePciReg 0x80000368, 0x00000066, 0xFFFFFF00     ;68h:66
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7 
    jmp codeend 
  .sixtyeighth8:
    writePciReg 0x80000368, 0x00000077, 0xFFFFFF00     ;68h:77
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7   
    jmp codeend 
  .sixtyeighth9:
    writePciReg 0x80000368, 0x00000088, 0xFFFFFF00     ;68h:88
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7   
    jmp codeend 
  .sixtyeighth10:
    writePciReg 0x80000368, 0x00000099, 0xFFFFFF00     ;68h:99
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7 
    jmp codeend 
  .sixtyeighth11:
    writePciReg 0x80000368, 0x000000AA, 0xFFFFFF00     ;68h:AA

    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7  
    jmp codeend 
  .sixtyeighth12:
    writePciReg 0x80000368, 0x000000BB, 0xFFFFFF00     ;68h:BB
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7  
    jmp codeend 
  .sixtyeighth13:
    writePciReg 0x80000368, 0x000000CC, 0xFFFFFF00     ;68h:CC
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7  
    jmp codeend 
  .sixtyeighth14:
    writePciReg 0x80000368, 0x000000DD, 0xFFFFFF00     ;68h:DD
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7  
    jmp codeend 
  .sixtyeighth15:
    writePciReg 0x80000368, 0x000000EE, 0xFFFFFF00     ;68h:EE
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7   
    jmp codeend 
  .sixtyeighth16:
    writePciReg 0x80000368, 0x000000FF, 0xFFFFFF00     ;68h:FF
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7 
    jmp codeend
  .sixtyeighth0:
    xor eax, eax
    readCmosReg 0x1B, 0xFF                             ; read FSB value from CMOS 1B register write to al
    shr eax, 4                                         ; shift al by 4 bits lower
    cmp al, 0x01
    je .fsb1
    cmp al, 0x02
    je .fsb2
    cmp al, 0x03
    je .fsb3
    cmp al, 0x04
    je .fsb4
    cmp al, 0x05
    je .fsb5
    cmp al, 0x06
    je .fsb6
    cmp al, 0x07
    je .fsb7
    cmp al, 0x08
    je .fsb8
    cmp al, 0x09
    je .fsb9
    cmp al, 0x0A
    je .fsbA
    cmp al, 0x0B
    je .fsbB
;
    .fsb1:
    .fsb2:
    .fsb3:
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7
    writePciReg 0x80000368, 0x000000FF, 0xFFFFFF00     ;68h:FF   
    jmp codeend
    .fsb4:
    .fsb5:
    .fsb6:
    writePciReg 0x80000364, 0x00F70000, 0xFF00FFFF     ;66h:F7
    writePciReg 0x80000368, 0x000000DD, 0xFFFFFF00     ;68h:DD
    jmp codeend
    .fsb7:
    writePciReg 0x80000364, 0x00880000, 0xFF00FFFF     ;66h:88
    writePciReg 0x80000368, 0x00000088, 0xFFFFFF00     ;68h:88
    jmp codeend
    .fsb8:
    writePciReg 0x80000364, 0x00550000, 0xFF00FFFF     ;66h:55
    writePciReg 0x80000368, 0x00000055, 0xFFFFFF00     ;68h:55   
    jmp codeend 
    .fsb9:
    writePciReg 0x80000364, 0x00330000, 0xFF00FFFF     ;66h:33
    writePciReg 0x80000368, 0x00000033, 0xFFFFFF00     ;68h:33
    jmp codeend
    .fsbA:
    .fsbB:
    writePciReg 0x80000364, 0x00110000, 0xFF00FFFF     ;66h:11
    writePciReg 0x80000368, 0x00000011, 0xFFFFFF00     ;68h:11   
    jmp codeend 
;
;
;
;====================================================================

codeend:
    popa
    popfd
    retf                                        ; return far to system bios routine

                                        
  times (ROM_SIZE_IN_BYTE-$) db 0               ; use 00h as the padding bytes until we reach the ROM size
                                                ; The last byte (512th) will be the patch_byte for the checksum
                                                ; patch_byte is calculated and automagically inserted below
  PREV_CHKSUM = 0
  repeat $
  load CHKSUM byte from %-1
  CHKSUM = (PREV_CHKSUM + CHKSUM) mod 0x100
  PREV_CHKSUM = CHKSUM
  end repeat
  
  store byte (0x100 - CHKSUM) at ($-1)          ; store the patch_byte
ROMEnd:
